#include "stdafx.h"
#include "AsyncSocketEx_bsdsocket.h"

#include <errno.h>
#include <iostream>
#include <vector>
#include <arpa/inet.h>
#include "IGTPoller.h"
#include <netdb.h>
#include "GTStopWatch.h"
#include "GTLog.h"
#include "GTCommonAPI.h"
#include <netinet/tcp.h> /* for TCP_NODELAY */
#include <unistd.h>

//#define __ALWAYS_SECOND_THREAD_MODE__

static unsigned long g_nPresetAsyncSocket = 0;

CAsyncSocketEx::CAsyncSocketEx() : m_hSocket(INVALID_SOCKET)
	, m_bRequestConnect(false)
	, m_bIsTcpListenSocket(false)
    , m_bFirstPollAction(true)	
	, m_bDeleteFlag(false)
#ifdef __ALWAYS_SECOND_THREAD_MODE__
    , m_bMainThreadMode(false)
#else
    , m_bMainThreadMode(true)
#endif
	, m_bSendingFlag(false)
	, m_bRecvingFlag(true)
	, m_nRetainCount(0)
{	
	_Retain();
	
	m_pEventDelegate = CreateEventDelegate(this);
	m_pEventDelegate->RegisterObserver(this);
}


CAsyncSocketEx::~CAsyncSocketEx()
{
	DLOG("[ENTER] CAsyncSocketEx::~CAsyncSocketEx() %p %d %p delegate = %p\n", this, m_hSocket, pthread_self(), m_pEventDelegate);

	ASSERT(m_bDeleteFlag);
	
	if(m_hSocket != INVALID_SOCKET)
	{
		GTPollerAPI::GetGTPollerInstance()->del(m_hSocket, this);

		close(m_hSocket);
		shutdown(m_hSocket, 2);
		
		m_hSocket = INVALID_SOCKET;
	}
	
	m_pEventDelegate->Delete();
	m_pEventDelegate = NULL;
	
	DLOG("[LEAVE] CAsyncSocketEx::~CAsyncSocketEx() %p %d %p\n", this, m_hSocket, pthread_self());
}

bool CAsyncSocketEx::Create(unsigned int nSocketPort, int nSocketType)
{
	if(SOCK_STREAM == nSocketType)
		m_hSocket = ::socket(AF_INET, nSocketType, IPPROTO_TCP);
	else
		m_hSocket = ::socket(AF_INET, nSocketType, IPPROTO_UDP);
	m_nSocketType = nSocketType;
	
	if(m_hSocket != INVALID_SOCKET && SOCK_DGRAM == nSocketType)	// if Only UDP, add to kqueue
	{
		GTPollerAPI::SetNonblocking(m_hSocket);
		
		GTPollerAPI::GetGTPollerInstance()->add(m_hSocket, this);
	}

#warning "[ENGINE] TCP NO DELAY OPTION SELECT"
	socklen_t onoff = GTCommonAPI::GTSetting::ReadIntSettingBundle("tcpNoDelay");
	if(setsockopt(m_hSocket, IPPROTO_TCP, TCP_NODELAY, (void *)&onoff, sizeof(onoff)) < 0)
	{
		GT_PLOG("TCP_NODELAY option error! : onoff = %d", onoff);
	}
		
	return (m_hSocket != INVALID_SOCKET);
}
	

// This Function is NOT THREAD SAFE!!!
void CAsyncSocketEx::Delete()
{
	if(false == m_bDeleteFlag)  // don't need to check.... twice delete() call is critical problem..
	{
		m_bDeleteFlag = true;
		
		//------------------------
		// All Resouce release..
		//------------------------
		
		// 1. GT Poller Client Callback
		Close();
		
		// 2. Event Delete Callback
		m_pEventDelegate->UnRegisterObserver(this);
		m_pEventDelegate->ReleaseAllSourcesAndTimers();		
		
		_Release();
	}
}


bool CAsyncSocketEx::GetPeerName( sockaddr* lpSockAddr, int* lpSockAddrLen )
{
	if ( !getpeername(m_hSocket, lpSockAddr, (socklen_t *)lpSockAddrLen) )
		return true;
	else
		return false;
}

bool CAsyncSocketEx::GetSockName( sockaddr* lpSockAddr, int* lpSockAddrLen )
{
	if ( !getsockname(m_hSocket, lpSockAddr, (socklen_t *)lpSockAddrLen) )
		return TRUE;
	else
		return FALSE;
}

bool CAsyncSocketEx::Accept( CAsyncSocketEx& rConnectedSocket, SOCKADDR* lpSockAddr, int* lpSockAddrLen )
{
	USES_SOCKET_AUTO_RELEASE;

	struct sockaddr_in sin;
	socklen_t len = sizeof(sin);
	
	int hNewSocket = accept(m_hSocket, (struct sockaddr *)&sin, &len);
	if(-1 == hNewSocket)
		return false;
	
	rConnectedSocket.m_bRequestConnect = true;
	rConnectedSocket.m_hSocket = hNewSocket;
	rConnectedSocket.m_bFirstPollAction = false;	

	GTPollerAPI::SetNonblocking(hNewSocket);

	GTPollerAPI::GetGTPollerInstance()->add(hNewSocket, &rConnectedSocket);

	return true;
}

bool CAsyncSocketEx::Bind(unsigned int nSocketPort, const char* lpszSocketAddress)
{
	SOCKADDR_IN sockAddr;
	memset(&sockAddr,0,sizeof(sockAddr));
		
	sockAddr.sin_family = AF_INET;
	
	if (lpszSocketAddress == NULL)
		sockAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	else
	{
		DWORD lResult = inet_addr(lpszSocketAddress);
		if (lResult == INADDR_NONE)
		{
			errno = EINVAL;
			return false;
		}
		sockAddr.sin_addr.s_addr = lResult;
	}
	
	sockAddr.sin_port = htons((u_short)nSocketPort);
	
	return Bind((SOCKADDR*)&sockAddr, sizeof(sockAddr));	
}

bool CAsyncSocketEx::Bind(const sockaddr* lpSockAddr, int nSockAddrLen)
{
	if (!bind(m_hSocket, lpSockAddr, nSockAddrLen))
		return true;
	else
		return false;

	return false;
}

bool CAsyncSocketEx::Listen( int nConnectionBacklog )
{
	bool bRes = false;
	bRes = (SOCKET_ERROR != listen(m_hSocket, nConnectionBacklog));
	
	if( bRes )
	{
		m_bIsTcpListenSocket = true;
		
		GTPollerAPI::SetNonblocking(m_hSocket);
		
		GTPollerAPI::GetGTPollerInstance()->add(m_hSocket, this);
	}
	return true;
}

bool CAsyncSocketEx::ShutDown( int nHow /*=sends*/ )
{
	if(m_hSocket >= 0)
	{
		if(!shutdown(m_hSocket, nHow))
			return true;
		
		return false;
	}
}

void CAsyncSocketEx::Close()
{
	if(m_hSocket != INVALID_SOCKET)
	{
		DLOG("CAsyncSocketEx::Close() this = %p, fd = %d, thread = %p\n", this, m_hSocket, pthread_self()); fflush(stdout);

		GTPollerAPI::GetGTPollerInstance()->del(m_hSocket, this);

		shutdown(m_hSocket, 2);
		close(m_hSocket);
				
		m_hSocket = INVALID_SOCKET;
	}
	
	m_bIsTcpListenSocket = false;
//	m_bRequestConnect = false;
}

int CAsyncSocketEx::GetLastError() 
{ 
	return errno;
}


void CAsyncSocketEx::Attach(int hSocket)
{
	USES_SOCKET_AUTO_RELEASE;

	m_hSocket = hSocket;

	GTPollerAPI::SetNonblocking(m_hSocket);
	
	GTPollerAPI::GetGTPollerInstance()->add(m_hSocket, this);
}


int CAsyncSocketEx::Detach()
{
	int socket = m_hSocket;
	
	GTPollerAPI::GetGTPollerInstance()->del(m_hSocket, this);

	m_hSocket = INVALID_SOCKET;
	
	return socket;
}


int CAsyncSocketEx::GetSocketHandle(void)
{
	return m_hSocket;
}

void CAsyncSocketEx::SetMainThreadEventMode(bool bEnable)
{
#ifdef __ALWAYS_SECOND_THREAD_MODE__ 
    return;
#else    
	m_bMainThreadMode = bEnable;	
	
	CheckOnRecvEvent();
#endif 
}

bool CAsyncSocketEx::Connect(const char* lpszHostAddress, unsigned int nHostPort, int nTimeOut )
{
	USES_SOCKET_AUTO_RELEASE;
	
	m_bRequestConnect = true;
	
	GT_STOP_WATCH_START(g_nPresetAsyncSocket, "CONNECT");
	
	std::string strIP;
	struct sockaddr_in sockaddr;
	sockaddr.sin_family = AF_INET;
	sockaddr.sin_port = htons(nHostPort);	
	sockaddr.sin_addr.s_addr = ::inet_addr(lpszHostAddress);

	if(sockaddr.sin_addr.s_addr == INADDR_NONE)
	{
		m_strHostName = lpszHostAddress;
		m_nCurrentPort = nHostPort;
		m_nCurrentTimeout = nTimeOut;
		
		char result[50] = {0, };
#ifdef __USE_ASYNC_HOSTBYNAME__
		if(AsyncHostByName::GetIPAddressFromHostCache(m_strHostName.c_str(), result, sizeof(result)) == false)
		{
			// Request resolving host..
			if(AsyncHostByName::GetHostByNameAsync(lpszHostAddress, this) == true)
            {
                _Retain();
                return true;
            }
			return false;
		}
#else
		hostent * pHostent = NULL;
		pHostent = gethostbyname(m_strHostName.c_str());
		if( pHostent )
            strcpy( result, inet_ntoa( *(in_addr*)pHostent->h_addr ) ); 
		else
			return false;
#endif
		sockaddr.sin_addr.s_addr = ::inet_addr(result);
	}
	
	m_bFirstPollAction = true;
	
	//------------------------------------------
	// NON BLOCKING SOCKET CONNECT
	//------------------------------------------
	GTPollerAPI::SetNonblocking(m_hSocket);
	
	int ret = ::connect( m_hSocket, (SOCKADDR*) &sockaddr, sizeof(sockaddr));

	if(ret != 0 && (ret != -1 || EINPROGRESS != errno))
	{
		Close();
		
		return false;
	}

	GTPollerAPI::GetGTPollerInstance()->add(m_hSocket, this);

	CheckOnSendEvent();	
	
	return true;
}
	
int CAsyncSocketEx::Receive(void* lpBuf, int nBufLen)
{
	if( NULL == lpBuf || 0 >= nBufLen )
		return -1;
	
	int ret = recv(m_hSocket, lpBuf, nBufLen, 0);
	return ret;
}

int CAsyncSocketEx::ReceiveFrom(void* lpBuf, int nBufLen, SOCKADDR* lpSockAddr, socklen_t* lpSockAddrLen, int nFlags )
{
	int ret = recvfrom(m_hSocket, lpBuf, nBufLen, nFlags, lpSockAddr, lpSockAddrLen);
	return ret;
}
	
int CAsyncSocketEx::Send(const void* lpBuf, int nBufLen)
{
	int ret = send(m_hSocket, lpBuf, nBufLen, 0);
	
	return ret;	
}

int CAsyncSocketEx::SendTo(const void* lpBuf, int nBufLen, const SOCKADDR* lpSockAddr, int nSockAddrLen, int nFlags )
{
	return sendto(m_hSocket, lpBuf, nBufLen, nFlags, lpSockAddr, nSockAddrLen);
}

void CAsyncSocketEx::UnCheckOnRecvEvent()
{
	if ( true == m_bRecvingFlag && m_hSocket > 0 )
	{
		GTPollerAPI::GetGTPollerInstance()->delFilter(m_hSocket, GTPollerReadFilter, this);
		m_bRecvingFlag = false;
	}
}


void CAsyncSocketEx::CheckOnRecvEvent()
{
	if ( false == m_bRecvingFlag && m_hSocket > 0)
	{
		GTPollerAPI::GetGTPollerInstance()->addFilter(m_hSocket, GTPollerReadFilter, this);
		m_bRecvingFlag = true;
	}
}

void CAsyncSocketEx::UnCheckOnSendEvent()
{
	if ( true == m_bSendingFlag && m_hSocket > 0)
	{
		GTPollerAPI::GetGTPollerInstance()->delFilter(m_hSocket, GTPollerWriteFilter, this);
		m_bSendingFlag = false;
	}
}


void CAsyncSocketEx::CheckOnSendEvent()
{
	if ( false == m_bSendingFlag && m_hSocket > 0)
	{
		GTPollerAPI::GetGTPollerInstance()->addFilter(m_hSocket, GTPollerWriteFilter, this);
		m_bSendingFlag = true;
	}
}

#ifdef __USE_ASYNC_HOSTBYNAME__
#pragma mark IGTHostByNameAsync

void CAsyncSocketEx::OnGetHostByNameAsyncResult(const char *szHostName, int nCount, void *userinfo)
{
    do {
        if(nCount <= 0) // RESOLVING FAILED..
        {
            OnConnect(-1);
            break;
        }
        
        char result[50] = {0, };
        if(AsyncHostByName::GetIPAddressFromHostCache(szHostName, result, sizeof(result)) == false)
        {
            ASSERT(false);	//<-- FATAL ERROR!! NONSENSE CASE..
            OnConnect(-2);
            break;
        }
		
        if(!Connect(result, m_nCurrentPort, m_nCurrentTimeout))
        {
            OnConnect(-3);
            break;
        }
        
    } while (false);
    
    //printf(">>>>>>>>>>>> OnGetHostByNameAsyncResult %p %d\n", szHostName, nCount);
    _Release();
}
#endif

//
//
//#pragma mark IGTEventDelegateEvent
//
//// IGTEventDelegateProtocol
//void CAsyncSocketEx::OnGTEventDelegateEvent_CanDelete(IGTEventDelegate *delegate, bool &deleteFlag)
//{
//	if(_Delete())
//		deleteFlag = true;
//	else
//		deleteFlag = false;
//}


#pragma mark IGTRunLoopObserver

bool CAsyncSocketEx::OnGTRunLoopObserver_SourceEvent(GT_EVENT_HANDLE handle, GT_PARAM param1, GT_PARAM param2)
{
//	if(m_bDeleteFlag)
//	{
//		GT_PLOG("!!!!!!!!!! CAsyncSocketEx::OnGTRunLoopObserver_SourceEvent : m_bDeleteFlag == true : %p", this);
//
//		return false;
//	}
	
	USES_SOCKET_AUTO_RELEASE;
	
//	if(INVALID_SOCKET == m_hSocket)
//	{
//		return false;
//	}
	
	//	printf("OnGTRunLoopObserver_SourceEvent %p %d %d %d\n", this, handle, param1, param2); fflush(stdout);	

	switch (handle) 
	{
		case WM_SOCKET_EVENT_ERROR:
			OnSocketError( (int)param1 );
			break;
			
		case WM_SOCKET_EVENT_RECEIVE:
			OnReceive( (int)param1, (int)param2 );
			
			CheckOnRecvEvent();
			break;
			
		case WM_SOCKET_EVENT_SEND:
			OnSend( (int)param1 );
			
			if(HasSendData())
				CheckOnSendEvent();
			break;

		case WM_SOCKET_EVENT_CONNECT:
			OnConnect( (int)param1 );
			break;

		case WM_SOCKET_EVENT_ACCEPT:
			OnAccept( (int)param1 );
			break;

		case WM_SOCKET_EVENT_CLOSE:			
			OnClose( (int)param1 );
			break;
		default:
			break;
	}
	
	return false;
}

#pragma mark IGTEventDelegateEvent

void CAsyncSocketEx::OnGTEventDelegateEvent_DidCreated(IGTEventDelegate *delegate)
{
	_Retain();
}


void CAsyncSocketEx::OnGTEventDelegateEvent_DidReleased(IGTEventDelegate *delegate)
{
	//GT_PLOG(">> !! CAsyncSocketEx::OnGTEventDelegateEvent_DidReleased %p retain : %d", this, m_nRetainCount);

	_Release();
}



#pragma mark GTPollerClient

void CAsyncSocketEx::OnGTPollerClient_RegisterEvent(IGTPoller *poller)
{
	//GT_PLOG(">> !! CAsyncSocketEx::OnGTPollerClient_RegisterEvent %p retain : %d", this, m_nRetainCount);
	_Retain();
}


void CAsyncSocketEx::OnGTPollerClient_UnregisterEvent(IGTPoller *poller)
{
	//GT_PLOG(">> !! CAsyncSocketEx::OnGTPollerClient_UnregisterEvent %p retain : %d", this, m_nRetainCount);
	_Release();	
}

void CAsyncSocketEx::OnGTPollerClient_NotifyPollEvent(GTPollerEvent *e)
{
//	printf("OnGTPollerClient_NotifyPollEvent %p filter : %d fd : %d %dbytes\n", this, e->filter, e->fd, e->bytes); fflush(stdout);	
	
//	if(m_bDeleteFlag)
//	{
//		GT_PLOG("!!!!!!!!!! CAsyncSocketEx::OnGTPollerClient_NotifyPollEvent : m_bDeleteFlag == true : %p", this);
//		return;
//	}
	
	USES_SOCKET_AUTO_RELEASE;

	do 
	{
		if(INVALID_SOCKET == m_hSocket)
		{
			// internal error..
			break;
		}
		
		if(e->flags & GTPOLLER_ERROR)
		{
			Fire_OnSocketError(-1);	// DO U KNOW HOW TO GET DETAIL ERRORCODE IN THIS CASE?? IF YES, TELL ME THAT...
			break;
		}
		
		if(e->flags & GTPOLLER_READ)
		{
			if(false == m_bIsTcpListenSocket)
			{
				Fire_OnReceive(0, e->bytes);
			}
			else
			{
				if((e->flags & GTPOLLER_EOF))
					Fire_OnClose(-1);
				else
					Fire_OnAccept(0);
			}
		}
		else if(e->flags & GTPOLLER_WRITE)
		{	
			if(m_bFirstPollAction)
			{
				m_bFirstPollAction = false;
				
				int error;
				socklen_t len = sizeof(error);
				if(getsockopt(e->fd, SOL_SOCKET, SO_ERROR, &error, &len) < 0 || error != 0)
				{
#ifdef __USE_ASYNC_HOSTBYNAME__
					AsyncHostByName::DeleteHostAddressCache(m_strHostName.c_str());
#endif
					UnCheckOnSendEvent();
					
					Close();
					Fire_OnConnect(-1);
					break;
				}
				GT_STOP_WATCH(g_nPresetAsyncSocket, "CONNECT");
				Fire_OnConnect(0);
			}
			
//			if(INVALID_SOCKET == m_hSocket)
//			{
//				break;
//			}
			
			if(!(e->flags & GTPOLLER_EOF))
				Fire_OnSend(0);	
			else
			{
				Fire_OnClose(-1);
			}
		}
	} while (false);
	
	// NEVER WRITE CODE HERE... CAN BE CRASHED...
	// ------------------------------------------
}
	
